/* char_lcd.c: A Linux Character-LCD driver for W5300E01-ARM board. */
/*
	Copyright 2008 WIZnet Co.,Ltd.

	This software may be used and distributed according to the terms of
	the GNU General Public License (GPL), incorporated herein by reference.
	Drivers based on or derived from this code fall under the GPL and must
	retain the authorship, copyright and license notice.  This file is not
	a complete program and may only be used when the entire operating
	system is licensed under the GPL.
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>

#include "char_lcd.h"
#define BUSY		0x80

#define DEV_NAME    "char_lcd"
#define DRV_VERSION "1.0"
#define DRV_RELDATE "June 20, 2008"

static const char driver_info[] =
KERN_INFO DEV_NAME ": WIZnet Character-LCD driver v" DRV_VERSION "(" DRV_RELDATE ")\n";

MODULE_AUTHOR("Taehun Kim <thkim@wiznet.co.kr>");
MODULE_DESCRIPTION("W5300E01-ARM Character-LCD driver");
MODULE_VERSION(DRV_VERSION);
MODULE_LICENSE("GPL");

char lcd_ready(void)
{
  //u8 flag;
#if 0
  volatile u8 flag;

  while( (flag=LcdCommandR) & BUSY)  
  {
    if(flag == 0xFF) return -1;  
  }
#endif
  mdelay(2);
  return 1;
}

void lcd_clrscr(void)       
{
  lcd_ready();
  LcdCommandW = 0x01;
}

char lcd_init(void)
{
  if((char)-1 ==lcd_ready()) return 0;

  LcdCommandW = 0x38;
  mdelay(50);
  LcdCommandW = 0x0C;
  mdelay(50);

  lcd_clrscr();
  lcd_gotoxy(0,0);
  return 1;
}

void lcd_gotoxy(u8 x,	u8 y)
{
  lcd_ready();
  switch(y)
  {
    case 0 : LcdCommandW = 0x80 + x; break;
    case 1 : LcdCommandW = 0xC0 + x; break;
    case 2 : LcdCommandW = 0x94 + x; break;
    case 3 : LcdCommandW = 0xD4 + x; break;
  }
}

char * lcd_puts(char* str)
{
  unsigned char i;

  for (i=0; str[i] != '\0'; i++){
    lcd_ready();
    LcdDataW = str[i];
  }
  return str;
}

void lcd_putch(char ch)
{
  lcd_ready();
  LcdDataW = ch;
}

void lcd_set_cursor_type(u8 type)	
{
  lcd_ready();
  switch(type) 
  {
    //No Cursor 
    case 0 : LcdCommandW = 0x0C; break;  
             // Normal Cursor 
    case 1 : LcdCommandW = 0x0E; break; 
             // No Cursor | Blink
    case 2 : LcdCommandW = 0x0D; break; 
             // Normal Cursor | Blink 
    case 3 : LcdCommandW = 0x0F; break; 
  }
}

void ShiftCursor(u8 dir)	
{
  lcd_ready();
  if(dir)
    LcdCommandW = 0x14;
  else
    LcdCommandW = 0x10;
}

void lcd_home_cursor(void)       
{
  lcd_ready();
  LcdCommandW = 0x2;
}

int __init 
char_lcd_module_init(void)
{
  printk(driver_info);
  lcd_init();
  lcd_puts("W5300E01-ARM");
  return 0;
}

void __exit
char_lcd_module_exit(void)
{
}

module_init(char_lcd_module_init);
module_exit(char_lcd_module_exit);
