#ifndef _WIZSOCK_H_
#define _WIZSOCK_H_
#include <net/sock.h> 
#include <asm/semaphore.h>
#include <linux/inetdevice.h>

#include "w5300.h"

/* UDP Header Information */
struct wiz_udphdr {
	u32 addr;
	u16 port;
	u16 size;
};

/* wiz_sock to be used for H/W TCP/IP */
struct wiz_sock {
	struct sock sk;
	int s; /* Channel information related to a socket */
  int remain_pkt;
};

/* Function to covert struct sock into wiz_sock */
static inline struct wiz_sock *wiz_sk(const struct sock *sk)
{
  return (struct wiz_sock*)sk;
}

/* Function to configure address information
 * Configuring IP address and net mask into W5300 by acquiring from Linux kernel */
static inline int set_addr_info(struct sock *sk)
{
  struct in_device *in_dev;
  struct in_ifaddr **ifap = NULL;
  struct net_device *dev = wiz_desc->dev;
  struct in_ifaddr *ifa = NULL;
  int found=0;

  /* Finding network interface of w5300 */
  if ((in_dev = __in_dev_get_rtnl(dev)) != NULL) {
			for (ifap = &in_dev->ifa_list; (ifa = *ifap) != NULL
          ;ifap = &ifa->ifa_next)
      {
				if (!strcmp("wiz0", ifa->ifa_label))
        { 
          /* found */
          found=1;
					break;
        }
      }
	}

  if(found)
  {
    /* Configuring IP address of w5300  */
    w5300_set_ipaddr(htonl(ifa->ifa_local));
    /* Configuring Net Mask of w5300 */
    w5300_set_netmask(htonl(ifa->ifa_mask));
    return 0;
  }

  /* Network interface of w5300 is not found */
  printk(KERN_ERR "WIZnet network interface not found!\n");
  return -EINVAL;
}
#endif /* _WIZSOCK_H_ */
